from polars import Expr

from polars_ta.wq.arithmetic import log_diff
from polars_ta.wq.time_series import ts_returns

# 对数收益
log_return = log_diff
# 简单收益
percent_return = ts_returns


def cum_return(close: Expr) -> Expr:
    """累计收益"""

    return close / close.drop_nulls().first()



import numpy as np
import polars as pl
import talib

from polars_ta.utils.helper import TaLibHelper

_ = TaLibHelper


class TestDemoClass:
    high_np = None
    low_np = None
    close_np = None
    df_pl = None

    def setup_class(self):
        self.high_np = np.arange(100, dtype=float) + np.random.rand(100)
        self.low_np = np.arange(100, dtype=float) - np.random.rand(100)
        self.close_np = np.arange(100, dtype=float)

        self.df_pl = pl.DataFrame([self.high_np, self.low_np, self.close_np],
                                  schema=["high", "low", "close"])

    def test_SMA(self):
        result1 = talib.SMA(self.close_np, timeperiod=5)
        result2 = self.df_pl.select(pl.col('close').ta.SMA(timeperiod=5))
        result3 = result2['close'].to_numpy()
        # print()
        # print(result1)
        # print(result3)

        assert np.allclose(result1, result3, equal_nan=True)

    def test_BBANDS(self):
        result1, _, _ = talib.BBANDS(self.close_np, timeperiod=5)
        result2 = self.df_pl.select(
            pl.col('close').ta.BBANDS(timeperiod=5, schema=['upperband', 'middleband', 'lowerband'], skip_nan=False).alias('BBANDS')
        ).unnest('BBANDS')
        result3 = result2['upperband'].to_numpy()
        # print()
        # print(result1)
        # print(result3)

        assert np.allclose(result1, result3, equal_nan=True)



from polars import Boolean, Int32, Series
from polars import Expr
from polars import when

from polars_ta.ta.overlap import EMA as _ema
from polars_ta.ta.overlap import SMA as MA
from polars_ta.ta.volatility import TRANGE as TR  # noqa
from polars_ta.wq.arithmetic import max_ as MAX  # noqa
from polars_ta.wq.arithmetic import min_ as MIN  # noqa
from polars_ta.wq.cross_sectional import rank as RANK  # noqa
from polars_ta.wq.time_series import ts_arg_max as HHVBARS  # noqa
from polars_ta.wq.time_series import ts_arg_min as LLVBARS  # noqa
from polars_ta.wq.time_series import ts_count as COUNT  # noqa
from polars_ta.wq.time_series import ts_decay_linear as WMA  # noqa
from polars_ta.wq.time_series import ts_delay as REF  # noqa
from polars_ta.wq.time_series import ts_delta as DIFF  # noqa
from polars_ta.wq.time_series import ts_max as HHV  # noqa
from polars_ta.wq.time_series import ts_min as LLV  # noqa
from polars_ta.wq.time_series import ts_product as MULAR  # noqa
from polars_ta.wq.time_series import ts_sum as SUM


def BARSLAST(condition: Expr) -> Expr:
    """BARSLAST(X),上一次X不为0到现在的天数"""
    a = condition.cum_count()
    b = when(condition.cast(Boolean)).then(a).otherwise(None).forward_fill()
    return a - b


def BARSLASTCOUNT(condition: Expr) -> Expr:
    """BARSLASTCOUNT(X),统计连续满足条件的周期数"""
    a = condition.cast(Int32).cum_sum()
    b = when(condition.cast(Boolean)).then(None).otherwise(a).forward_fill().fill_null(0)
    return a - b


def BARSSINCE(condition: Expr) -> Expr:
    """BARSSINCE(X):第一次X不为0到现在的天数"""
    a = condition.cum_count()
    b = condition.cast(Boolean).arg_true().first()
    return a - b


def _bars_since_n(x: Series) -> Series:
    a = x.cast(Boolean).arg_true()
    # 返回的值可能为空，所以需要判断一下
    b = a[0] if len(a) > 0 else float('nan')
    return len(x) - 1 - b


def BARSSINCEN(condition: Expr, timeperiod: int = 30) -> Expr:
    """BARSSINCEN(X,N):N周期内第一次X不为0到现在的天数"""
    return condition.rolling_map(_bars_since_n, timeperiod)


def DMA(close: Expr, alpha: float = 0.5) -> Expr:
    """DMA(X,A),求X的动态移动平均.
算法:Y=A*X+(1-A)*Y',其中Y'表示上一周期Y值,A必须大于0且小于1.A支持变量"""
    return close.ewm_mean(alpha=alpha, adjust=False, min_periods=1)


def EMA(close: Expr, timeperiod: int = 30) -> Expr:
    """EMA(X,N):X的N日指数移动平均.算法:Y=(X*2+Y'*(N-1))/(N+1)
 EMA(X,N)相当于SMA(X,N+1,2),N支持变量"""
    return _ema(close, timeperiod)


def EXPMA(close: Expr, timeperiod: int = 30) -> Expr:
    return _ema(close, timeperiod)


def EXPMEMA(close: Expr, timeperiod: int = 30) -> Expr:
    """EXPMEMA(X,M),X的M日指数平滑移动平均。EXPMEMA同EMA(即EXPMA)的差别在于他的起始值为一平滑值

    Notes
    -----
    等价于talib.EMA，由于比EMA慢，少用

    """
    sma = MA(close, timeperiod)
    x = when(close.cum_count() < timeperiod).then(sma).otherwise(close)
    return x.ewm_mean(span=timeperiod, adjust=False, min_periods=1)


def _hod(x: Series):
    return x.rank(descending=True)[-1]


def HOD(close: Expr, timeperiod: int = 30) -> Expr:
    """HOD(X,N):求当前X数据是N周期内的第几个高值,N=0则从第一个有效值开始"""
    return close.rolling_map(_hod, timeperiod)


def _lod(x: Series):
    return x.rank(descending=False)[-1]


def LOD(close: Expr, timeperiod: int = 30) -> Expr:
    """LOD(X,N):求当前X数据是N周期内的第几个低值"""
    return close.rolling_map(_lod, timeperiod)


def MEMA(close: Expr, timeperiod: int = 30) -> Expr:
    """MEMA(X,N):X的N日平滑移动平均,如Y=(X+Y'*(N-1))/N
 MEMA(X,N)相当于SMA(X,N,1)"""
    raise


def RANGE(a: Expr, b: Expr, c: Expr) -> Expr:
    """A在B和C范围之间,B<A<C."""
    return (b < a) & (a < c)


def SMA(X: Expr, N: int, M: int = 1) -> Expr:
    """用法:SMA(X,N,M),X的N日移动平均,M为权重,若Y=SMA(X,N,M)则Y=(X*M+Y'*(N-M))/N"""
    return X.ewm_mean(alpha=M / N, adjust=False, min_periods=1)


def SUM_0(close: Expr) -> Expr:
    return close.cum_sum()


def SUMIF(condition: Expr, close: Expr, timeperiod: int = 30) -> Expr:
    return SUM(condition.cast(Int32) * close, timeperiod)


def TMA(close: Expr, timeperiod: int = 30) -> Expr:
    """TMA(X,A,B),A和B必须小于1,算法	Y=(A*Y'+B*X),其中Y'表示上一周期Y值.初值为X"""
    raise


def FILTER(close: Expr, timeperiod: int = 30) -> Expr:
    raise



import numpy as np
from polars import Expr, Series, rolling_corr, Int32, rolling_cov, arange, repeat

from polars_ta.utils.pandas_ import roll_kurt
from polars_ta.utils.pandas_ import roll_rank


# TODO rolling_map比较慢，少用. 如ts_arg_max、ts_product等

def _arg_max(x: Series):
    """
    Notes
    -----
    等polars推出rolling_arg_max(reverse=True)这个问题能好转

    """
    # return x[::-1].arg_max()
    # return x.reverse().arg_max() # 正确，但太慢
    return len(x) - 1 - x.arg_max()  # 有多个最大值相同时，靠前的值会被记录下来，导致结果偏大


def ts_arg_max(x: Expr, d: int = 5) -> Expr:
    # WorldQuant中最大值为今天返回0，为昨天返回1
    return x.rolling_map(_arg_max, d)


def _arg_min(x: Series):
    return len(x) - 1 - x.arg_min()


def ts_arg_min(x: Expr, d: int = 5) -> Expr:
    return x.rolling_map(_arg_min, d)


def ts_co_kurtosis(x: Expr, y: Expr, d: int = 5, ddof: int = 1) -> Expr:
    raise


def ts_co_skewness(x: Expr, y: Expr, d: int = 5, ddof: int = 1) -> Expr:
    raise


def ts_corr(x: Expr, y: Expr, d: int = 5, ddof: int = 1) -> Expr:
    # x、y不区分先后
    return rolling_corr(x, y, window_size=d, ddof=ddof)


def ts_count(x: Expr, d: int = 30) -> Expr:
    return x.cast(Int32).rolling_sum(d)


def ts_count_nans(x: Expr, d: int = 5) -> Expr:
    # null与nan到底用哪一个？
    return x.is_null().rolling_sum(d)


def ts_covariance(x: Expr, y: Expr, d: int = 5, ddof: int = 1) -> Expr:
    # x、y不区分先后
    return rolling_cov(x, y, window_size=d, ddof=ddof)


def ts_decay_exp_window(x: Expr, d: int = 30, factor: float = 1.0) -> Expr:
    # TODO weights not yet supported on array with null values
    y = arange(d - 1, -1, step=-1, eager=False)
    weights = repeat(factor, d, eager=True).pow(y)
    return x.rolling_mean(d, weights=weights)


def ts_decay_linear(x: Expr, d: int = 30, dense: bool = False) -> Expr:
    # TODO weights not yet supported on array with null values
    weights = arange(1, d + 1, eager=True)
    return x.rolling_mean(d, weights=weights)


def ts_delay(x: Expr, d: int = 1) -> Expr:
    return x.shift(d)


def ts_delta(x: Expr, d: int = 1) -> Expr:
    return x.diff(d)


def ts_ir(x: Expr, d: int = 1) -> Expr:
    return ts_mean(x, d) / ts_std_dev(x, d)


def ts_kurtosis(x: Expr, d: int = 5) -> Expr:
    # TODO 等待polars官方出rolling_kurt
    return x.map_batches(lambda a: roll_kurt(a, d))


def ts_max(x: Expr, d: int = 30) -> Expr:
    return x.rolling_max(d)


def ts_max_diff(x: Expr, d: int = 30) -> Expr:
    return x - ts_max(x, d)


def ts_mean(x: Expr, d: int = 5) -> Expr:
    return x.rolling_mean(d)


def ts_median(x: Expr, d: int = 5) -> Expr:
    return x.rolling_median(d)


def ts_min(x: Expr, d: int = 30) -> Expr:
    return x.rolling_min(d)


def ts_min_diff(x: Expr, d: int = 30) -> Expr:
    return x - ts_min(x, d)


def ts_product(x: Expr, d: int = 5) -> Expr:
    return x.rolling_map(np.nanprod, d)


def ts_rank(x: Expr, d: int = 5, constant=0) -> Expr:
    # TODO 等待polars官方出rolling_rank，并支持pct
    # bottleneck长期无人维护，pydata/bottleneck#434 没有合并
    # pandas中已经用跳表实现了此功能，速度也不差
    return x.map_batches(lambda a: roll_rank(a, d, True))


def ts_returns(x: Expr, d: int = 1) -> Expr:
    return x.pct_change(d)


def ts_scale(x: Expr, d: int = 5) -> Expr:
    a = ts_min(x, d)
    b = ts_max(x, d)
    return (x - a) / (b - a)


def ts_skewness(x: Expr, d: int = 5, bias=False) -> Expr:
    # bias=False与pandas结果一样
    return x.rolling_skew(d, bias=bias)


def ts_std_dev(x: Expr, d: int = 5, ddof: int = 0) -> Expr:
    return x.rolling_std(d, ddof=ddof)


def ts_sum(x: Expr, d: int = 30) -> Expr:
    return x.rolling_sum(d)


def ts_weighted_delay(x: Expr, k: float = 0.5) -> Expr:
    return x.rolling_sum(2, weights=[1 - k, k])


def ts_zscore(x: Expr, d: int = 5) -> Expr:
    return (x - ts_mean(x, d)) / ts_std_dev(x, d)



"""
本文件中参数全用
open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr
open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr
统一的好处是在使用时不用考虑函数调用区别
"""
from polars import Expr

from polars_ta import TA_EPSILON
from polars_ta.candles.cdl1 import four_price_doji, dragonfly, gravestone


def limit_up(price: Expr, high_limit: Expr) -> Expr:
    """
    开盘 涨停
    收盘 涨停
    最高 曾涨停
    """
    return price >= high_limit - TA_EPSILON


def limit_up_at_open(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """开盘 涨停"""
    return limit_up(open_, high_limit)


def limit_up_at_close(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """收盘 涨停"""
    return limit_up(close, high_limit)


def limit_up_at_high(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """曾涨停"""
    return limit_up(high, high_limit) & ~limit_up(close, high_limit)


def limit_up_four_price_doji(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """一字涨停"""
    return four_price_doji(open_, high, low, close) & limit_up(close, high_limit)


def limit_up_dragonfly(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """T字涨停"""
    return limit_up(close, high_limit) & dragonfly(open_, high, low, close)


# ======================================
def limit_down(price: Expr, low_limit: Expr) -> Expr:
    """
    开盘 跌停
    收盘 跌停
    最低 曾跌停
    """
    return price <= low_limit + TA_EPSILON


def limit_down_at_open(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """开盘 跌停"""
    return limit_down(open_, low_limit)


def limit_down_at_close(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """收盘 跌停"""
    return limit_down(close, low_limit)


def limit_down_at_high(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """曾跌停"""
    return limit_down(low, low_limit) & ~limit_down(close, low_limit)


def limit_down_four_price_doji(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """一字跌停"""
    return four_price_doji(open_, high, low, close) & limit_down(close, low_limit)


def limit_down_gravestone(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """T字跌停"""
    return limit_down(open_, low_limit) & gravestone(open_, high, low, close)



from typing import Sequence

import numpy as np
from polars import Expr, Series


def standardize_zscore(x: Expr, ddof: int = 0) -> Expr:
    return (x - x.mean()) / x.std(ddof=ddof)


def standardize_minmax(x: Expr) -> Expr:
    a = x.min()
    b = x.max()
    return (x - a) / (b - a)


def winsorize_quantile(x: Expr, low_limit: float = 0.025, up_limit: float = 0.995) -> Expr:
    a = x.quantile(low_limit)
    b = x.quantile(up_limit)
    return x.clip(lower_bound=a, upper_bound=b)


def winsorize_3sigma(x: Expr, n: float = 3.) -> Expr:
    a = x.mean()
    b = n * x.std(ddof=0)
    return x.clip(lower_bound=a - b, upper_bound=a + b)


def winsorize_mad(x: Expr, n: float = 3., k: float = 1.4826) -> Expr:
    # https://en.wikipedia.org/wiki/Median_absolute_deviation
    a = x.median()
    b = (n * k) * (x - a).abs().median()
    return x.clip(lower_bound=a - b, upper_bound=a + b)


def neutralize_demean(x: Expr) -> Expr:
    return x - x.mean()


def neutralize_residual_simple(y: Expr, x: Expr) -> Expr:
    # https://stackoverflow.com/a/74906705/1894479
    # 一元回归时，这个版本更快，不需再补充常量1
    # e_i = y_i - a - bx_i
    #     = y_i - ȳ + bx̄ - bx_i
    #     = y_i - ȳ - b(x_i - x̄)
    x_demeaned = x - x.mean()
    y_demeaned = y - y.mean()
    x_demeaned_squared = x_demeaned.pow(2)
    beta = x_demeaned.dot(y_demeaned) / x_demeaned_squared.sum()
    return y_demeaned - beta * x_demeaned


def neutralize_residual_multiple(cols: Sequence[Series]) -> Series:
    # https://stackoverflow.com/a/74906705/1894479
    # 比struct.unnest要快一些
    cols = [c.to_numpy() for c in cols]
    y = cols[0]
    x = cols[1:]
    A = np.vstack(x).T
    coef = np.linalg.lstsq(A, y, rcond=None)[0]
    y_hat = np.sum(A * coef, axis=1)
    residual = y - y_hat
    return Series(residual)



from polars import Boolean
from polars import Expr
from polars import when

from polars_ta.wq.logical import if_else


def IF(condition: Expr, a: Expr, b: Expr) -> Expr:
    """IF(X,A,B)若X不为0则返回A,否则返回B"""
    return if_else(condition.cast(Boolean), a, b)


def IFN(condition: Expr, a: Expr, b: Expr) -> Expr:
    """IFN(X,A,B)若X不为0则返回B,否则返回A"""
    return if_else(condition.cast(Boolean), b, a)


def VALUEWHEN(condition: Expr, x: Expr) -> Expr:
    return when(condition).then(x).otherwise(None).forward_fill()


IFF = IF



from polars import Expr


def AD(high: Expr, low: Expr, close: Expr, volume: Expr) -> Expr:
    """"""
    raise


def ADOSC(high: Expr, low: Expr, close: Expr, volume: Expr, fastperiod: int = 3, slowperiod: int = 10) -> Expr:
    """"""
    raise


def OBV(close: Expr, volume: Expr) -> Expr:
    """"""
    raise



# generated by codegen_talib.py
import talib as _ta
from polars import Expr, struct

from polars_ta.utils.helper import TaLibHelper, func_wrap_11, func_wrap_mn

_ = TaLibHelper


def HT_DCPERIOD(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.HT_DCPERIOD, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def HT_DCPHASE(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.HT_DCPHASE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def HT_PHASOR(close: Expr, skip_nan=False, output_idx=None, schema=('inphase', 'quadrature'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.HT_PHASOR, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def HT_SINE(close: Expr, skip_nan=False, output_idx=None, schema=('sine', 'leadsine'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.HT_SINE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def HT_TRENDMODE(close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.HT_TRENDMODE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ADD(high: Expr, low: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.ADD, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def DIV(high: Expr, low: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.DIV, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MAX(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.MAX, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MAXINDEX(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.MAXINDEX, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MIN(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.MIN, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MININDEX(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.MININDEX, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MINMAX(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('min', 'max'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.MINMAX, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MINMAXINDEX(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('minidx', 'maxidx'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.MINMAXINDEX, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MULT(high: Expr, low: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.MULT, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def SUB(high: Expr, low: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.SUB, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def SUM(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.SUM, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ACOS(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.ACOS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ASIN(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.ASIN, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ATAN(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.ATAN, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CEIL(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.CEIL, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def COS(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.COS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def COSH(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.COSH, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def EXP(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.EXP, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def FLOOR(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.FLOOR, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def LN(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.LN, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def LOG10(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.LOG10, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def SIN(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.SIN, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def SINH(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.SINH, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def SQRT(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.SQRT, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def TAN(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.TAN, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def TANH(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.TANH, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ADX(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.ADX, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ADXR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.ADXR, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def APO(close: Expr, fastperiod: int = 12, slowperiod: int = 26, matype: int = 0, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.APO, x, fastperiod, slowperiod, matype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def AROON(high: Expr, low: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('aroondown', 'aroonup'), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.AROON, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def AROONOSC(high: Expr, low: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.AROONOSC, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def BOP(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.BOP, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CCI(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CCI, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CMO(close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.CMO, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def DX(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.DX, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MACD(close: Expr, fastperiod: int = 12, slowperiod: int = 26, signalperiod: int = 9, skip_nan=False, output_idx=None, schema=('macd', 'macdsignal', 'macdhist'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.MACD, x, fastperiod, slowperiod, signalperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MACDEXT(close: Expr, fastperiod: int = 12, fastmatype: int = 0, slowperiod: int = 26, slowmatype: int = 0, signalperiod: int = 9, signalmatype: int = 0, skip_nan=False, output_idx=None, schema=('macd', 'macdsignal', 'macdhist'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.MACDEXT, x, fastperiod, fastmatype, slowperiod, slowmatype, signalperiod, signalmatype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MACDFIX(close: Expr, signalperiod: int = 9, skip_nan=False, output_idx=None, schema=('macd', 'macdsignal', 'macdhist'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.MACDFIX, x, signalperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MFI(high: Expr, low: Expr, close: Expr, volume: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close, volume).map_batches(lambda x: func_wrap_mn(_ta.MFI, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MINUS_DI(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.MINUS_DI, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MINUS_DM(high: Expr, low: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.MINUS_DM, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MOM(close: Expr, timeperiod: int = 10, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.MOM, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def PLUS_DI(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.PLUS_DI, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def PLUS_DM(high: Expr, low: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.PLUS_DM, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def PPO(close: Expr, fastperiod: int = 12, slowperiod: int = 26, matype: int = 0, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.PPO, x, fastperiod, slowperiod, matype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ROC(close: Expr, timeperiod: int = 10, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.ROC, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ROCP(close: Expr, timeperiod: int = 10, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.ROCP, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ROCR(close: Expr, timeperiod: int = 10, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.ROCR, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ROCR100(close: Expr, timeperiod: int = 10, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.ROCR100, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def RSI(close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.RSI, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def STOCH(high: Expr, low: Expr, close: Expr, fastk_period: int = 5, slowk_period: int = 3, slowk_matype: int = 0, slowd_period: int = 3, slowd_matype: int = 0, skip_nan=False, output_idx=None, schema=('slowk', 'slowd'), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.STOCH, x, fastk_period, slowk_period, slowk_matype, slowd_period, slowd_matype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def STOCHF(high: Expr, low: Expr, close: Expr, fastk_period: int = 5, fastd_period: int = 3, fastd_matype: int = 0, skip_nan=False, output_idx=None, schema=('fastk', 'fastd'), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.STOCHF, x, fastk_period, fastd_period, fastd_matype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def STOCHRSI(close: Expr, timeperiod: int = 14, fastk_period: int = 5, fastd_period: int = 3, fastd_matype: int = 0, skip_nan=False, output_idx=None, schema=('fastk', 'fastd'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.STOCHRSI, x, timeperiod, fastk_period, fastd_period, fastd_matype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def TRIX(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.TRIX, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ULTOSC(high: Expr, low: Expr, close: Expr, timeperiod1: int = 7, timeperiod2: int = 14, timeperiod3: int = 28, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.ULTOSC, x, timeperiod1, timeperiod2, timeperiod3, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def WILLR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.WILLR, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def BBANDS(close: Expr, timeperiod: int = 5, nbdevup: float = 2.0, nbdevdn: float = 2.0, matype: int = 0, skip_nan=False, output_idx=None, schema=('upperband', 'middleband', 'lowerband'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.BBANDS, x, timeperiod, nbdevup, nbdevdn, matype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def DEMA(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.DEMA, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def EMA(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.EMA, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def HT_TRENDLINE(close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.HT_TRENDLINE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def KAMA(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.KAMA, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MA(close: Expr, timeperiod: int = 30, matype: int = 0, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.MA, x, timeperiod, matype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MAMA(close: Expr, fastlimit: float = 0.5, slowlimit: float = 0.05, skip_nan=False, output_idx=None, schema=('mama', 'fama'), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_mn(_ta.MAMA, x, fastlimit, slowlimit, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MAVP(close: Expr, periods: Expr, minperiod: int = 2, maxperiod: int = 30, matype: int = 0, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(close, periods).map_batches(lambda x: func_wrap_mn(_ta.MAVP, x, minperiod, maxperiod, matype, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MIDPOINT(close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.MIDPOINT, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MIDPRICE(high: Expr, low: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.MIDPRICE, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def SAR(high: Expr, low: Expr, acceleration: float = 0.02, maximum: float = 0.2, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.SAR, x, acceleration, maximum, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def SAREXT(high: Expr, low: Expr, startvalue: float = 0.0, offsetonreverse: float = 0.0, accelerationinitlong: float = 0.02, accelerationlong: float = 0.02, accelerationmaxlong: float = 0.2, accelerationinitshort: float = 0.02, accelerationshort: float = 0.02, accelerationmaxshort: float = 0.2, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.SAREXT, x, startvalue, offsetonreverse, accelerationinitlong, accelerationlong, accelerationmaxlong, accelerationinitshort, accelerationshort, accelerationmaxshort, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def SMA(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.SMA, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def T3(close: Expr, timeperiod: int = 5, vfactor: float = 0.7, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.T3, x, timeperiod, vfactor, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def TEMA(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.TEMA, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def TRIMA(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.TRIMA, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def WMA(close: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.WMA, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDL2CROWS(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDL2CROWS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDL3BLACKCROWS(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDL3BLACKCROWS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDL3INSIDE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDL3INSIDE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDL3LINESTRIKE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDL3LINESTRIKE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDL3OUTSIDE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDL3OUTSIDE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDL3STARSINSOUTH(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDL3STARSINSOUTH, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDL3WHITESOLDIERS(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDL3WHITESOLDIERS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLABANDONEDBABY(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLABANDONEDBABY, x, penetration, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLADVANCEBLOCK(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLADVANCEBLOCK, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLBELTHOLD(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLBELTHOLD, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLBREAKAWAY(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLBREAKAWAY, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLCLOSINGMARUBOZU(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLCLOSINGMARUBOZU, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLCONCEALBABYSWALL(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLCONCEALBABYSWALL, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLCOUNTERATTACK(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLCOUNTERATTACK, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLDARKCLOUDCOVER(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.5, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLDARKCLOUDCOVER, x, penetration, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLDOJI(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLDOJI, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLDOJISTAR(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLDOJISTAR, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLDRAGONFLYDOJI(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLDRAGONFLYDOJI, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLENGULFING(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLENGULFING, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLEVENINGDOJISTAR(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLEVENINGDOJISTAR, x, penetration, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLEVENINGSTAR(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLEVENINGSTAR, x, penetration, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLGAPSIDESIDEWHITE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLGAPSIDESIDEWHITE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLGRAVESTONEDOJI(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLGRAVESTONEDOJI, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLHAMMER(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLHAMMER, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLHANGINGMAN(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLHANGINGMAN, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLHARAMI(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLHARAMI, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLHARAMICROSS(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLHARAMICROSS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLHIGHWAVE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLHIGHWAVE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLHIKKAKE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLHIKKAKE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLHIKKAKEMOD(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLHIKKAKEMOD, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLHOMINGPIGEON(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLHOMINGPIGEON, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLIDENTICAL3CROWS(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLIDENTICAL3CROWS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLINNECK(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLINNECK, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLINVERTEDHAMMER(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLINVERTEDHAMMER, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLKICKING(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLKICKING, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLKICKINGBYLENGTH(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLKICKINGBYLENGTH, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLLADDERBOTTOM(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLLADDERBOTTOM, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLLONGLEGGEDDOJI(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLLONGLEGGEDDOJI, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLLONGLINE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLLONGLINE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLMARUBOZU(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLMARUBOZU, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLMATCHINGLOW(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLMATCHINGLOW, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLMATHOLD(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.5, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLMATHOLD, x, penetration, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLMORNINGDOJISTAR(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLMORNINGDOJISTAR, x, penetration, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLMORNINGSTAR(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLMORNINGSTAR, x, penetration, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLONNECK(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLONNECK, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLPIERCING(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLPIERCING, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLRICKSHAWMAN(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLRICKSHAWMAN, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLRISEFALL3METHODS(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLRISEFALL3METHODS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLSEPARATINGLINES(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLSEPARATINGLINES, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLSHOOTINGSTAR(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLSHOOTINGSTAR, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLSHORTLINE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLSHORTLINE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLSPINNINGTOP(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLSPINNINGTOP, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLSTALLEDPATTERN(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLSTALLEDPATTERN, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLSTICKSANDWICH(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLSTICKSANDWICH, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLTAKURI(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLTAKURI, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLTASUKIGAP(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLTASUKIGAP, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLTHRUSTING(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLTHRUSTING, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLTRISTAR(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLTRISTAR, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLUNIQUE3RIVER(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLUNIQUE3RIVER, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLUPSIDEGAP2CROWS(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLUPSIDEGAP2CROWS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CDLXSIDEGAP3METHODS(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('integer',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.CDLXSIDEGAP3METHODS, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def AVGPRICE(open: Expr, high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(open, high, low, close).map_batches(lambda x: func_wrap_mn(_ta.AVGPRICE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def MEDPRICE(high: Expr, low: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.MEDPRICE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def TYPPRICE(high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.TYPPRICE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def WCLPRICE(high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.WCLPRICE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def BETA(high: Expr, low: Expr, timeperiod: int = 5, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.BETA, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def CORREL(high: Expr, low: Expr, timeperiod: int = 30, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low).map_batches(lambda x: func_wrap_mn(_ta.CORREL, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def LINEARREG(close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.LINEARREG, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def LINEARREG_ANGLE(close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.LINEARREG_ANGLE, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def LINEARREG_INTERCEPT(close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.LINEARREG_INTERCEPT, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def LINEARREG_SLOPE(close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.LINEARREG_SLOPE, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def STDDEV(close: Expr, timeperiod: int = 5, nbdev: float = 1.0, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.STDDEV, x, timeperiod, nbdev, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def TSF(close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.TSF, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def VAR(close: Expr, timeperiod: int = 5, nbdev: float = 1.0, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return close.map_batches(lambda x: func_wrap_11(_ta.VAR, x, timeperiod, nbdev, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ATR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.ATR, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def NATR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.NATR, x, timeperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def TRANGE(high: Expr, low: Expr, close: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close).map_batches(lambda x: func_wrap_mn(_ta.TRANGE, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def AD(high: Expr, low: Expr, close: Expr, volume: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close, volume).map_batches(lambda x: func_wrap_mn(_ta.AD, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def ADOSC(high: Expr, low: Expr, close: Expr, volume: Expr, fastperiod: int = 3, slowperiod: int = 10, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(high, low, close, volume).map_batches(lambda x: func_wrap_mn(_ta.ADOSC, x, fastperiod, slowperiod, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))


def OBV(close: Expr, volume: Expr, skip_nan=False, output_idx=None, schema=('real',), schema_format='{}', nan_to_null=False) -> Expr:
    return struct(close, volume).map_batches(lambda x: func_wrap_mn(_ta.OBV, x, skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null))




from polars import Expr

from polars_ta.wq.preprocess import neutralize_demean
from polars_ta.wq.preprocess import standardize_minmax
from polars_ta.wq.preprocess import standardize_zscore
from polars_ta.wq.preprocess import winsorize_3sigma


# 函数名都没有加`cs_`前缀，使用时需留心

def normalize(x: Expr, use_std: bool = False, limit: float = 0.0) -> Expr:
    """Calculates the mean value of all valid alpha values for a certain date, then subtracts that mean from each element."""
    if use_std:
        # 这里用ddof=1才能与文档示例的数值对应上
        r = standardize_zscore(x, ddof=1)
    else:
        r = neutralize_demean(x)

    if limit == 0:
        return r
    else:
        return r.clip(-limit, limit)


def one_side(x: Expr, is_long: bool = True) -> Expr:
    """Shifts all instruments up or down so that the Alpha becomes long-only or short-only
(if side = short), respectively."""
    # TODO: 这里不确定，需再研究
    # [-1, 0, 1]+1=[0, 1, 2]
    # max([-1, 0, 1], 0)=[0,0,1]
    raise


def rank(x: Expr, rate: int = 2, pct: bool = True) -> Expr:
    """Ranks the input among all the instruments and returns an equally distributed number between 0.0 and 1.0. For precise sort, use the rate as 0."""
    if pct:
        # TODO 等待count的变更
        return x.rank() / (x.count() - x.null_count())
    else:
        return x.rank()


def scale(x: Expr, scale_=1, long_scale=1, short_scale=1) -> Expr:
    """Scales input to booksize. We can also scale the long positions and short positions to separate scales by mentioning additional parameters to the operator."""
    if long_scale != 1 or short_scale != 1:
        L = x.clip(lower_bound=0)  # 全正数
        S = x.clip(upper_bound=0)  # 全负数，和还是负数
        return L / L.sum() * long_scale - S / S.sum() * short_scale
    else:
        return x / x.abs().sum() * scale_


def scale_down(x: Expr, constant: float = 0) -> Expr:
    """Scales all values in each day proportionately between 0 and 1 such that minimum value maps to 0 and maximum value maps to 1. Constant is the offset by which final result is subtracted."""
    return standardize_minmax(x)


def truncate(x: Expr, max_percent: float = 0.01) -> Expr:
    """Operator truncates all values of x to maxPercent. Here, maxPercent is in decimal notation."""
    return x.clip(upper_bound=x.sum() * max_percent)


def winsorize(x: Expr, std: float = 4) -> Expr:
    return winsorize_3sigma(x, std)


def zscore(x: Expr) -> Expr:
    return standardize_zscore(x)



from polars import Expr, when


# arccos等函数与tanh放在transformational.py而不是arithmetic.py下可能是把它们看成是激活函数

def arc_cos(x: Expr) -> Expr:
    """If -1 <= x <= 1: arccos(x); else NaN"""
    return x.arccos()


def arc_sin(x: Expr) -> Expr:
    """If -1 <= x <= 1: arcsin(x); else NaN"""
    return x.arcsin()


def arc_tan(x: Expr) -> Expr:
    """This operator does inverse tangent of input. """
    return x.arctan()


def bucket(x: Expr,
           range="0, 1, 0.1",
           buckets="2,5,6,7,10",
           skipBegin=False, skipEnd=False, skipBoth=False,
           NANGroup=True) -> Expr:
    """Convert float values into indexes for user-specified buckets. Bucket is useful for creating group values, which can be passed to group operators as input."""
    # TODO 未完成
    raise


def clamp(x: Expr, lower: float = 0, upper: float = 0, inverse: bool = False, mask=None) -> Expr:
    """Limits input value between lower and upper bound in inverse = false mode (which is default). Alternatively, when inverse = true, values between bounds are replaced with mask, while values outside bounds are left as is."""
    if inverse:
        # mask is one of: 'nearest_bound', 'mean', 'NAN' or any floating point number
        return when((lower < x) & (x < upper)).then(mask).otherwise(x)
    else:
        return x.clip(lower, upper)


def filter_(x, h="1, 2, 3, 4", t="0.5"):
    """Used to filter the value and allows to create filters like linear or exponential decay."""
    raise


def keep(x, f, period=5):
    """This operator outputs value x when f changes and continues to do that for “period” days after f stopped changing. After “period” days since last change of f, NaN is output."""
    raise


def left_tail(x: Expr, maximum: float = 0) -> Expr:
    """NaN everything greater than maximum, maximum should be constant."""
    return when(x > maximum).then(None).otherwise(x)


def pasteurize(x: Expr) -> Expr:
    """Set to NaN if x is INF or if the underlying instrument is not in the Alpha universe"""
    # TODO: 不在票池中的的功能无法表示
    # TODO: 与purify好像没啥区别
    return when(x.is_infinite()).then(None).otherwise(x)


def right_tail(x: Expr, minimum: float = 0) -> Expr:
    """NaN everything less than minimum, minimum should be constant."""
    return when(x < minimum).then(None).otherwise(x)


def sigmoid(x: Expr) -> Expr:
    """Returns 1 / (1 + exp(-x))"""
    return 1 / (1 + (-x).exp())


def tail(x: Expr, lower: float = 0, upper: float = 0, newval: float = 0) -> Expr:
    """If (x > lower AND x < upper) return newval, else return x. Lower, upper, newval should be constants. """
    # TODO 与clamp一样?
    return when((lower < x) & (x < upper)).then(newval).otherwise(x)


def tanh(x: Expr) -> Expr:
    """Hyperbolic tangent of x"""
    return x.tanh()



from polars import Expr
from polars import reduce, max_horizontal, sum_horizontal, min_horizontal, Int64
from polars import when


def abs_(x: Expr) -> Expr:
    return x.abs()


def add(*args, filter_=False):
    """Add all inputs (at least 2 inputs required). If filter = true, filter all input NaN to 0 before adding"""
    if filter_:
        # TODO 等官方修复此bug
        # https://github.com/pola-rs/polars/issues/13113
        # return sum_horizontal(*args)
        args = [_.fill_null(0) for _ in args]

    return reduce(function=lambda acc, x: acc + x, exprs=args)


def ceiling(x: Expr) -> Expr:
    return x.ceil()


def cos(x: Expr) -> Expr:
    return x.cos()


def cosh(x: Expr) -> Expr:
    return x.cosh()


def densify(x: Expr) -> Expr:
    raise


def divide(x: Expr, y: Expr) -> Expr:
    return x / y


def exp(x: Expr) -> Expr:
    return x.exp()


def floor(x: Expr) -> Expr:
    return x.floor()


def fraction(x: Expr) -> Expr:
    """This operator removes the whole number part and returns the remaining fraction part with sign."""
    # return sign(x) * (abs(x) - floor(abs(x)))
    # return x.sign() * (x.abs() % 1)
    return x % 1


def inverse(x: Expr) -> Expr:
    """1 / x"""
    return 1 / x


def log(x: Expr) -> Expr:
    return x.log()


def log10(x: Expr) -> Expr:
    return x.log10()


def log1p(x: Expr) -> Expr:
    return x.log1p()


def log_diff(x: Expr, d: int = 1) -> Expr:
    """Returns log(current value of input or x[t] ) - log(previous value of input or x[t-1])."""
    return x.log().diff(d)


def max_(*args):
    """Maximum value of all inputs. At least 2 inputs are required."""
    return max_horizontal(args)


def mean(*args) -> Expr:
    # TODO 还需要修正
    return sum_horizontal(*args) / len(args)


def min_(*args):
    """Maximum value of all inputs. At least 2 inputs are required."""
    return min_horizontal(args)


def mod(x: Expr, y: Expr) -> Expr:
    return x % y


def multiply(*args, filter_=False):
    """Multiply all inputs. At least 2 inputs are required. Filter sets the NaN values to 1"""
    if filter_:
        args = [_.fill_null(1) for _ in args]

    return reduce(function=lambda acc, x: acc * x, exprs=args)


def power(x: Expr, y: Expr) -> Expr:
    """x ^ y"""
    return x.pow(y)


def purify(x: Expr) -> Expr:
    """Clear infinities (+inf, -inf) by replacing with NaN."""
    return when(x.is_infinite()).then(None).otherwise(x)


def reverse(x: Expr) -> Expr:
    """- x"""
    return -x


def round_(x: Expr, decimals: int = 0) -> Expr:
    """Round input to closest integer."""
    return x.round(decimals)


def round_down(x: Expr, f: int = 1) -> Expr:
    """Round input to greatest multiple of f less than input;"""
    if f == 1:
        return x // 1
    else:
        return x // f * f


def s_log_1p(x: Expr) -> Expr:
    return (x.abs() + 1).log() * x.sign()


def sign(x: Expr) -> Expr:
    return x.sign()


def signed_power(x: Expr, y: Expr) -> Expr:
    """x raised to the power of y such that final result preserves sign of x."""
    if isinstance(y, (int, float)):
        if y == 1:
            return x.abs() * x.sign()
        elif y == 0:
            return x.sign()

    return x.abs().pow(y) * x.sign()


def sin(x: Expr) -> Expr:
    return x.sin()


def sinh(x: Expr) -> Expr:
    return x.sinh()


def sqrt(x: Expr) -> Expr:
    return x.sqrt()


def subtract(*args, filter_=False) -> Expr:
    """x-y. If filter = true, filter all input NaN to 0 before subtracting"""
    if filter_:
        args = [_.fill_null(0) for _ in args]

    return reduce(function=lambda acc, x: acc - x, exprs=args)


def tan(x: Expr) -> Expr:
    return x.tan()


def truncate(x: Expr) -> Expr:
    """向零取整"""
    return x.cast(Int64)



from polars import Expr
from polars import Series

from polars_ta.wq.time_series import ts_corr as RELATE  # noqa
from polars_ta.wq.time_series import ts_covariance as COVAR  # noqa
from polars_ta.wq.time_series import ts_std_dev


def _avedev(x: Series) -> Series:
    # 可惜rolling_map后这里已经由Expr变成了Series
    return (x - x.mean()).abs().mean()


def AVEDEV(close: Expr, timeperiod: int = 5) -> Expr:
    """平均绝对偏差"""
    return close.rolling_map(_avedev, timeperiod)


def DEVSQ(close: Expr, timeperiod: int = 5) -> Expr:
    raise


def SLOPE(close: Expr, timeperiod: int = 5) -> Expr:
    raise


def STD(close: Expr, timeperiod: int = 5) -> Expr:
    """估算标准差"""
    return ts_std_dev(close, timeperiod, 1)


def STDDEV(close: Expr, timeperiod: int = 5) -> Expr:
    """标准偏差?"""
    raise


def STDP(close: Expr, timeperiod: int = 5) -> Expr:
    """总体标准差"""
    return ts_std_dev(close, timeperiod, 0)


def VAR(close: Expr, timeperiod: int = 5) -> Expr:
    return close.rolling_var(timeperiod, ddof=1)


def VARP(close: Expr, timeperiod: int = 5) -> Expr:
    return close.rolling_var(timeperiod, ddof=0)
